from logging import getLogger

from gumo.core import EntityKey
from gumo.core import EntityKeyFactory

from gumo.datastore._configuration import configure
from gumo.datastore.infrastructure.configuration import DatastoreConfiguration

from gumo.datastore.infrastructure.repository import datastore_transaction


__all__ = [
    configure.__name__,

    EntityKey.__name__,
    EntityKeyFactory.__name__,
    DatastoreConfiguration.__name__,

    datastore_transaction.__name__,
]

logger = getLogger('gumo.datastore')


configure()



import os
import dataclasses
import threading
import requests

from typing import Optional
from typing import ClassVar
from typing import Union

from gumo.core import GoogleCloudProjectID

from google.cloud import datastore
import google.auth.credentials


class EmulatorCreds(google.auth.credentials.Credentials):
    def __init__(self):
        self.token = b'secret'
        self.expiry = None

    @property
    def valid(self):
        return True

    def refresh(self, _):
        raise RuntimeError('Should never be refreshed.')


@dataclasses.dataclass(frozen=False)
class DatastoreConfiguration:
    google_cloud_project: Union[GoogleCloudProjectID, str, None] = None
    use_local_emulator: bool = False
    emulator_host: Optional[str] = None
    namespace: Optional[str] = None
    client: Optional[datastore.Client] = None

    _ENV_KEY_GOOGLE_CLOUD_PROJECT: ClassVar = 'GOOGLE_CLOUD_PROJECT'
    _ENV_KEY_DATASTORE_EMULATOR_HOST: ClassVar = 'DATASTORE_EMULATOR_HOST'

    _lock: ClassVar = threading.Lock()

    def __post_init__(self):
        with self._lock:
            self._set_google_cloud_project()
            self._set_emulator_config()
            self._set_client()

    def _set_google_cloud_project(self):
        if isinstance(self.google_cloud_project, str):
            self.google_cloud_project = GoogleCloudProjectID(self.google_cloud_project)
        if isinstance(self.google_cloud_project, GoogleCloudProjectID):
            if self.google_cloud_project.value != os.environ.get(self._ENV_KEY_GOOGLE_CLOUD_PROJECT):
                raise RuntimeError(f'Env-var "{self._ENV_KEY_GOOGLE_CLOUD_PROJECT}" is invalid or undefined.'
                                   f'Please set value "{self.google_cloud_project.value}" to env-vars.')

        if self.google_cloud_project is None:
            if self._ENV_KEY_GOOGLE_CLOUD_PROJECT in os.environ:
                self.google_cloud_project = GoogleCloudProjectID(os.environ[self._ENV_KEY_GOOGLE_CLOUD_PROJECT])
            else:
                raise RuntimeError(f'Env-var "{self._ENV_KEY_GOOGLE_CLOUD_PROJECT}" is undefined, please set it.')

    def _set_emulator_config(self):
        emulator_not_configured = not self.use_local_emulator and self.emulator_host is None
        if emulator_not_configured and os.environ.get(self._ENV_KEY_DATASTORE_EMULATOR_HOST):
            self.use_local_emulator = True
            self.emulator_host = os.environ[self._ENV_KEY_DATASTORE_EMULATOR_HOST]

        if not self.use_local_emulator:
            return

        if os.environ.get(self._ENV_KEY_DATASTORE_EMULATOR_HOST) is None:
            raise RuntimeError(
                f'If the emulator enabled, then env-var "{self._ENV_KEY_DATASTORE_EMULATOR_HOST}" must be present.'
            )

        if os.environ.get(self._ENV_KEY_DATASTORE_EMULATOR_HOST) != self.emulator_host:
            host = os.environ.get(self._ENV_KEY_DATASTORE_EMULATOR_HOST)
            raise RuntimeError(
                f'Env-var "{self._ENV_KEY_DATASTORE_EMULATOR_HOST}" and self.emulator_host do not match. '
                f'env["{self._ENV_KEY_DATASTORE_EMULATOR_HOST}"]={host}, self.emulator_host={self.emulator_host}'
            )

    def _set_client(self):
        if isinstance(self.client, datastore.Client):
            return

        if self._ENV_KEY_DATASTORE_EMULATOR_HOST in os.environ:
            emulator_credentials = EmulatorCreds()
            self.client = datastore.Client(
                project=self.google_cloud_project.value,
                namespace=self.namespace,
                credentials=emulator_credentials,
                _http=requests.Session()
            )
            return

        self.client = datastore.Client(
            project=self.google_cloud_project.value,
            namespace=self.namespace,
        )



from gumo.datastore.infrastructure.repository import DatastoreRepositoryMixin


class DatastoreRepositoryMixinForTest(DatastoreRepositoryMixin):
    KIND = None

    def cleanup_entities(self):
        if self.KIND is None:
            raise RuntimeError('KIND must be present.')

        query = self.datastore_client.query(kind=self.KIND)
        query.keys_only()
        self.datastore_client.delete_multi(keys=[entity.key for entity in query.fetch()])

    def count_entities(self) -> int:
        if self.KIND is None:
            raise RuntimeError('KIND must be present.')

        query = self.datastore_client.query(kind=self.KIND)
        query.keys_only()
        return len(list(query.fetch()))



from logging import getLogger
from injector import singleton

from typing import Optional
from typing import Union

from gumo.core.injector import injector
from gumo.datastore.infrastructure.configuration import DatastoreConfiguration


logger = getLogger('gumo.datastore')


class ConfigurationFactory:
    @classmethod
    def build(
            cls,
            use_local_emulator: Union[str, bool, None] = None,
            emulator_host: Optional[str] = None,
            namespace: Optional[str] = None,
    ) -> DatastoreConfiguration:
        _use_emulator = False
        if isinstance(use_local_emulator, bool):
            _use_emulator = use_local_emulator
        elif isinstance(use_local_emulator, str):
            _use_emulator = use_local_emulator.lower() in ['true', 'yes']

        return DatastoreConfiguration(
            use_local_emulator=_use_emulator,
            emulator_host=emulator_host,
            namespace=namespace,
        )


def configure(
        use_local_emulator: Union[str, bool, None] = None,
        emulator_host: Optional[str] = None,
        namespace: Optional[str] = None,
) -> DatastoreConfiguration:
    config = ConfigurationFactory.build(
        use_local_emulator=use_local_emulator,
        emulator_host=emulator_host,
        namespace=namespace,
    )

    logger.debug(f'Gumo.Datastore is configured, config={config}')
    injector.binder.bind(DatastoreConfiguration, to=config, scope=singleton)

    from gumo.datastore._bind import datastore_binder
    injector.binder.install(datastore_binder)

    return config



from gumo.datastore.infrastructure.repository import DatastoreRepositoryMixin
from gumo.datastore.infrastructure.repository import datastore_transaction
from gumo.datastore.infrastructure.entity_key_mapper import EntityKeyMapper
from gumo.datastore.infrastructure.mapper import DatastoreMapperMixin


DatastoreEntity = DatastoreRepositoryMixin.DatastoreEntity


__all__ = [
    DatastoreRepositoryMixin.__name__,
    DatastoreEntity.__name__,
    datastore_transaction.__name__,
    EntityKeyMapper.__name__,
    DatastoreMapperMixin.__name__,
]



from typing import Optional
from typing import List
from injector import inject

from gumo.core.application.entity_key import KeyIDAllocator
from gumo.core.domain.entity_key import NoneKey
from gumo.core.domain.entity_key import EntityKey
from gumo.core.domain.entity_key import EntityKeyFactory
from gumo.core.domain.entity_key import IncompleteKey

from gumo.datastore.infrastructure.configuration import DatastoreConfiguration
from gumo.datastore.infrastructure import DatastoreRepositoryMixin

from google.cloud import datastore


class _IncompleteKeyMapper:
    @inject
    def __init__(
            self,
            datastore_configuration: DatastoreConfiguration,
            entity_key_factory: EntityKeyFactory,
    ):
        self._datastore_client = datastore_configuration.client
        self._entity_key_factory = entity_key_factory

    def to_datastore_key(self, incomplete_key: IncompleteKey) -> Optional[datastore.Key]:
        if incomplete_key is None or isinstance(incomplete_key, NoneKey):
            return None

        datastore_key = self._datastore_client.key(*incomplete_key.flat_pairs())

        return datastore_key

    def to_entity_key(self, datastore_key: Optional[datastore.Key]) -> EntityKey:
        if datastore_key is None:
            return NoneKey.get_instance()

        entity_key = self._entity_key_factory.build_from_pairs(pairs=datastore_key.path)
        return entity_key


class DatastoreKeyIDAllocator(KeyIDAllocator, DatastoreRepositoryMixin):
    ALLOCATE_BATCH_SIZE = 10

    @inject
    def __init__(
            self,
            key_mapper: _IncompleteKeyMapper,
    ):
        self.key_mapper = key_mapper

        self._cache = {}

    def _allocate_keys(
            self,
            incomplete_key: IncompleteKey,
            num_keys: int,
    ) -> List[EntityKey]:
        keys = self._cache.get(incomplete_key, [])

        if len(keys) < num_keys:
            if num_keys < self.ALLOCATE_BATCH_SIZE:
                allocate_nums = self.ALLOCATE_BATCH_SIZE
            else:
                allocate_nums = num_keys

            keys.extend(
                self._allocate_keys_without_cache(
                    incomplete_key=incomplete_key,
                    num_keys=allocate_nums
                )
            )

        allocated_keys = keys[0:num_keys]
        self._cache[incomplete_key] = keys[num_keys:]

        return allocated_keys

    def _allocate_keys_without_cache(
            self,
            incomplete_key: IncompleteKey,
            num_keys: int,
    ) -> List[EntityKey]:
        datastore_key = self.key_mapper.to_datastore_key(incomplete_key=incomplete_key)
        allocated_keys = self.datastore_client.allocate_ids(
            incomplete_key=datastore_key,
            num_ids=num_keys,
        )
        return [
            self.key_mapper.to_entity_key(datastore_key=key)
            for key in allocated_keys
        ]

    def allocate_keys(
            self,
            incomplete_key: IncompleteKey,
            num_keys: Optional[int] = None,
    ) -> List[EntityKey]:
        if num_keys is None:
            num_keys = self.ALLOCATE_BATCH_SIZE

        return self._allocate_keys(
            incomplete_key=incomplete_key,
            num_keys=num_keys,
        )

    def allocate(self, incomplete_key: IncompleteKey) -> EntityKey:
        return self.allocate_keys(
            incomplete_key=incomplete_key,
            num_keys=1,
        )[0]
